#ifndef zc030x_main_h
#define zc030x_main_h
/*
 * ZC030X experimental driver based on USB Skeleton Driver 0.7
 *
 * Authors: 
 *             Martin Braun (martinbraun at gmx.net)
 *             Arnaud Patard (thertp at users.sourceforge.net)
 *
 * code based on: 
 *        USB Skeleton Driver 0.7
 *        SPCA50x based camera driver 
 *        phillips webcam drivers (pwc) 
 *         
 * 
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/videodev.h>
#include <linux/usb.h>

#define DRIVER_DESC "ZC030X USB Camera Driver"
#define DRIVER_VERSION "0.01 experimental"


/*
 *      tq_scheduler disappeared @ lk-2.4.0-test12
 *      (right when <linux/sched.h> newly defined TQ_ACTIVE)
 *
 *      (snagged from linux/drivers/message/fusion/mptscsih.h)
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0) 
#define HAVE_TQ_SCHED 1

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) 
  #include <linux/sched.h>
  #ifdef TQ_ACTIVE
    #undef HAVE_TQ_SCHED
  #endif /* TQ_ACTIVE */
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) */

#ifdef HAVE_TQ_SCHED
#define SCHEDULE_TASK(x)                \
        /*MOD_INC_USE_COUNT*/;          \
        (x)->next = NULL;               \
        queue_task(x, &tq_scheduler)
#else /* HAVE_TQ_SCHED */
#define SCHEDULE_TASK(x)                \
        /*MOD_INC_USE_COUNT*/;          \
        if (schedule_task(x) == 0) {    \
                /*MOD_DEC_USE_COUNT*/;  \
                                   }
#endif /* HAVE_TQ_SCHED */
#endif 

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
    #include <linux/workqueue.h>
    #define tq_struct work_struct 
    #define INIT_TQUEUE INIT_WORK
    #define schedule_task schedule_work 
    #define SCHEDULE_TASK schedule_work
#else
    #include <linux/tqueue.h>
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  #define urb_t struct urb
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20) */

/* kernel debugging messages */

/*
 *  Configure debugging level...
 *
 * 0 = no debug messages
 * 1 = init/detection/unload and other significant messages,
 * 2 = some warning messages
 * 3 = config/control function calls
 * 4 = most function calls and data parsing messages
 * 5 = highly repetitive mesgs
 * NOTE: This should be changed to 0, 1, or 2 for production kernels
 *
 */
static int pdebug;

#define ZC030X_ENABLE_DEBUG  /*enable debugging messages */

#ifdef ZC030X_ENABLE_DEBUG
#  define PDEBUG(level, fmt, args...) \
if (pdebug >= level) info("[%s:%d] " fmt, __PRETTY_FUNCTION__, __LINE__ , ## args)
#else 
#  define PDEBUG(level, fmt, args...) do {} while(0)
#endif 

/* known cameras (vendor & product id's)			
 * add constants here for new supported usb devices, add them to device list
 * in .c file too
 */

#define USB_W300A_VENDOR_ID    0x055f
#define USB_W300A_PRODUCT_ID   0xd003

#define USB_WEBCAMNXPRO_VENDOR_ID    0x041e
#define USB_WEBCAMNXPRO_PRODUCT_ID   0x401e

#define USB_WEBCAMNX_VENDOR_ID   0x041e
#define USB_WEBCAMNX_PRODUCT_ID  0x401c

#define USB_WEBCAMMOBILE_VENDOR_ID	0x041e
#define USB_WEBCAMMOBILE_PRODUCT_ID	0x4017

#define USB_WEBCAMNOTEBOOK_VENDOR_ID	0x041e
#define USB_WEBCAMNOTEBOOK_PRODUCT_ID	0x401f

#define USB_GENIUSV2_VENDOR_ID    0x0458
#define USB_GENIUSV2_PRODUCT_ID   0x7007

#define USB_BENQCAM_VENDOR_ID	0x0ac8
#define USB_BENQCAM_PRODUCT_ID	0x0302

#define USB_ZSTARCORP_VENDOR_ID	    0x0ac8
#define USB_ZSTARCORP_PRODUCT_ID    0x301b


#define VIDEO_INTERFACE 0

#define DEFAULT_CONFIGURATION	1

/* altsettings, number specifies max packet size */
#define ALT_IF_SETTING_0 0 
#define ALT_IF_SETTING_128 1 
#define ALT_IF_SETTING_192 2 
#define ALT_IF_SETTING_256 3 
#define ALT_IF_SETTING_384 4 
#define ALT_IF_SETTING_512 5 
#define ALT_IF_SETTING_768 6
#define ALT_IF_SETTING_1023 7

/* Size and number of buffers for the ISO pipe. */
#define MAX_ISO_BUFS		2
#define ISO_FRAMES_PER_DESC	10
#define ISO_MAX_FRAME_SIZE	768
#define ISO_BUFFER_SIZE 	(ISO_FRAMES_PER_DESC * ISO_MAX_FRAME_SIZE)

#define ZC030X_NUMFRAMES 4

/* Video Size 640 x 480 x 4 bytes for RGB */
#define FRAME_HEADER_SIZE 20 
#define MAX_VIDEOFRAME_SIZE (640 * 480 * 4)
#define MAX_VIDEODATA_SIZE ((640 * 480 * 4) + 20) //TODO: calculate this

struct cam_list {
  int id;
  const char *name;
};

/* Include independant declarations */
#include "zc030x.h"


/* State machine for each frame in the frame buffer during capture */
enum {
        STATE_SCANNING,         /* Scanning for start */
        STATE_HEADER,           /* Parsing header */
        STATE_LINES,            /* Parsing lines */
};

/* Buffer states */
enum {
        BUF_NOT_ALLOCATED,
        BUF_ALLOCATED,
        BUF_PEND_DEALLOC,       /* spca50x->buf_timer is set */
};

/* The following structures were based on cpia.h. Why reinvent the wheel? :-) */
struct iso_buf
{
	void *data;
	int  length;
	int  read;
	struct urb *urb;
};

/* States for each frame buffer. */
enum { 
    FRAME_UNUSED,           /* Unused (no MCAPTURE) */
    FRAME_READY,            /* Ready to start grabbing */
    FRAME_GRABBING,         /* In the process of being grabbed into */
    FRAME_DONE,             /* Finished grabbing, but not been synced yet */
    FRAME_ERROR,            /* Something bad happened while processing */
    FRAME_ABORTING,         /* Aborting everything. Caused by hot unplugging.*/
};

/* Structure to hold frame info & data */
struct zc030x_frame {
    unsigned char *data;            /* Frame buffer */
    unsigned char *tmpbuffer;       /* temporary buffer zc030x->tmpbuffer need for decoding*/
    short *YBuffer;         /* Temporary jpeg decoder workspace */
    short *UBuffer;
    short *VBuffer;

    int seq;                /* Frame sequence number */

    int depth;              /* Bytes per pixel */
    int width;              /* Width application is expecting */
    int height;             /* Height */

    int hdrwidth;           /* Width the frame actually is */
    int hdrheight;          /* Height */

    unsigned int format;    /* Format asked by apps for this frame */
    int cameratype;         /* native in frame format */

    volatile int grabstate; /* State of grabbing */
    int scanstate;          /* State of scanning */

    long scanlength;        /* uncompressed, raw data length of frame */
    long bytes_read;        /* amount of scanlength that has been read from *data */

    wait_queue_head_t wq;   /* Processes waiting */

    int last_packet;        /* sequence number for last packet */
    unsigned char *highwater; /* used for debugging */
};


struct sensor_info
{
    int type;             /* identifies sensor */
    int maxwidth;         /* max/min sensor resolution */
    int maxheight;
    int minwidth;
    int minheight;
};

struct cam_params {
    __u8 BandWidth;
    __u8 Brightness;
    __u8 Contrast;
    __u8 Hue;
    __u8 Saturation;
    __u8 Sharpness;
    __u8 Gamma;
    __u8 WhiteBalance;
    __u8 BackLight;
    __u8 ColorEnable;
    __u8 FrameRate;
    __u8 Exposure;
    __u8 Gain;
    __u8 Mirror;
    __u8 Flicker;
    __u8 Default;
    __u8 ExposureAuto;
    __u8 WhiteBalanceAuto;
    __u8 GainAuto;
    __u8 BandWidthAuto;
    __u8 Vertical;
    __u8 Rgain;
    __u8 Ggain;
    __u8 Bgain;
};

/* Structure to hold all of our device specific stuff */
struct usb_zc030x
{
  /********* USB device management *****************/
  
  struct usb_device *udev;	/* save off the usb device pointer */
  char busy;			
  
  struct usb_interface_descriptor *iface_desc; /*  */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
 
  devfs_handle_t devfs;		/* TODO: merge own - devfs device node */
#endif
  int open;                     /* TODO: same ? - if the port is open or not */
  int present;                  /* if the device is not disconnected */
  
  unsigned char minor;		/* the starting minor number for this device */

  struct semaphore sem;		/* locks this structure */

  int prod_id;                  /* for use when need diferent functions */
  int vend_id;
  int alt;                      /* current alternate interface */
  int ctrlend;                  /* control endpoint */
  int isocend;                  /* isoc endpoint */
  int intrend;                  /* button interrupt endpoint */
  int vendpoint;		/* video isoc endpoint */
 
  struct urb *inturb;           /* urb for button interrupt */
  
  int vmax_packet_size;		/* USB maxpacket size */

  /******** Camera/Sensor specific parameters *******/
  
  int type_id;		        /* Video device type (-> supported_cams)*/
  int sensor_id;                /* Sensor type (-> supported_sensors) */
  struct cam_params params;     /* What we think the hardware is currently set to */
 
  /********* jpeg decoder workspace *****************/
  
  char  *tmpBuffer;
  
  /********* Framebuffer/sbuf management ***********/
  
  struct semaphore buf_lock;
  /* wait queue */
  wait_queue_head_t wq;   /* Processes waiting */
  
  struct iso_buf sbuf[MAX_ISO_BUFS]; /* isoc receiving buffers */
  int buf_state;
  int cursbuf;			/* current isoc receiving buffer */
  int streaming;		/* isoc data is streaming */  
  struct zc030x_frame frame[ZC030X_NUMFRAMES];

  /********* V4L device & buffers ******************/
  
  struct video_device *vdev;	/* video device (v4l) */
  char *fbuf;			/* Video buffer, contains captured frame data*/ 
  char *mmbuf;                  /* video mmap buffer */
  int curframe;                 /* frame buffer */
  int videoformat;
};

#include "sensors/SensorSig.h"




unsigned long kvirt_to_pa(unsigned long adr);
#endif


